/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_protocol.h"
#include "chirp_client.h"

#include "link.h"
#include "auth.h"
#include "domain_name_cache.h"
#include "full_io.h"
#include "macros.h"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/statfs.h>

#include <unistd.h>
#include <sys/errno.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <netinet/tcp.h>

#define BUFFER_SIZE 65536

static int get_stat_result( struct chirp_client *c, struct stat *buf, time_t stoptime );
static int get_statfs_result( struct chirp_client *c, struct statfs *buf, time_t stoptime );
static int get_result( struct chirp_client *c, time_t stoptime );
static int convert_result( int response );
static int simple_command(struct chirp_client *c, time_t stoptime, char const *fmt,...);
static void vsprintf_chirp(char *command,char const *fmt,va_list args);

static int global_serial = 0;

struct chirp_client {
	struct link *link;
	int broken;
	int serial;
};

struct chirp_client * chirp_client_connect_condor( time_t stoptime )
{
	FILE *file;
	int fields;
	int save_errno;
	struct chirp_client *client;
	char host[CHIRP_LINE_MAX];
	char hostport[CHIRP_LINE_MAX];
	char cookie[CHIRP_LINE_MAX];
	int port;
	int result;

	file = fopen("chirp.config","r");
	if(!file) return 0;

	fields = fscanf(file,"%s %d %s",host,&port,cookie);
	fclose(file);

	if(fields!=3) {
		errno = EINVAL;
		return 0;
	}

	sprintf(hostport,"%s:%d",host,port);

	client = chirp_client_connect(hostport,0,stoptime);
	if(!client) return 0;

	result = chirp_client_cookie(client,cookie,stoptime);
	if(result!=0) {
		save_errno = errno;
		chirp_client_disconnect(client);
		errno = save_errno;
		return 0;
	}

	return client;
}

struct chirp_client * chirp_client_connect( const char *hostport, int negotiate_auth, time_t stoptime )
{
	struct chirp_client *c;
	char addr[LINK_ADDRESS_MAX];
	char host[DOMAIN_NAME_MAX];
	int save_errno;
	int port;

	if(sscanf(hostport,"%[^:]:%d",host,&port)==2) {
		/* use the split host and port */
	} else {
		strcpy(host,hostport);
		port = CHIRP_PORT;
	}

	if(!domain_name_cache_lookup(host,addr)) {
		errno = ENOENT;
		return 0;
	}

	c = malloc(sizeof(*c));
	if(c) {
		c->link = link_connect(addr,port,stoptime);
		c->broken = 0;
		c->serial = global_serial++;
		if(c->link) {
			link_tune(c->link,LINK_TUNE_INTERACTIVE);
			if(negotiate_auth) {
				char *type, *subject;
				if(auth_assert(c->link,&type,&subject,stoptime)) {
					free(type);
					free(subject);
					return c;
				} else {
					chirp_client_disconnect(c);
					c = 0;
					if(time(0)>=stoptime) {
						errno = ECONNRESET;
					} else {
						errno = EPERM;
					}
				}
			} else {
				return c;
			}
		}
		save_errno = errno;
		free(c);
		errno = save_errno;
	}

	return 0;
}

void chirp_client_disconnect( struct chirp_client *c )
{
	link_close(c->link);
	free(c);
}

int chirp_client_serial( struct chirp_client *c )
{
	return c->serial;
}

int chirp_client_cookie( struct chirp_client *c, const char *cookie, time_t stoptime )
{
	return simple_command(c,stoptime,"cookie %s\n",cookie);
}

int chirp_client_login( struct chirp_client *c, const char *name, const char *password, time_t stoptime  )
{
	return simple_command(c,stoptime,"login %s %s\n",name,password);
}

int chirp_client_lookup( struct chirp_client *c, const char *logical_name, char **url, time_t stoptime  )
{
	int result;
	int actual;

	result = simple_command(c,stoptime,"lookup %s\n",logical_name);

	if(result>0) {
		*url = malloc(result);
		if(*url) {
			actual = link_read(c->link,*url,result,stoptime);
			if(actual!=result) {
				c->broken = 1;
				errno = ECONNRESET;
				return -1;
			}
		} else {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

int chirp_client_getdir( struct chirp_client *c, const char *path, chirp_dir_t callback, void *arg, time_t stoptime  )
{
	int result;
	char name[CHIRP_PATH_MAX];

	result = simple_command(c,stoptime,"getdir %s\n",path);
	if(result>=0) {
		while(1) {
			if(link_readline(c->link,name,sizeof(name),stoptime)) {
				if(name[0]) {
					callback(name,arg);
				} else {
					break;
				}
			} else {
				c->broken = 1;
				errno = ECONNRESET;
				return -1;
			}
		}
	}
	return result;

}

int chirp_client_open( struct chirp_client *c, const char *path, int flags, int mode, struct stat *buf, time_t stoptime  )
{
	int result;
	char fstr[256];
	fstr[0] = 0;

	if(flags&O_WRONLY) {
		strcat(fstr,"w");
	} else if(flags&O_RDWR) {
		strcat(fstr,"rw");
	} else {
		strcat(fstr,"r");
	}

	if(flags&O_CREAT) strcat(fstr,"c");
	if(flags&O_TRUNC) strcat(fstr,"t");
	if(flags&O_APPEND) strcat(fstr,"a");
	if(flags&O_SYNC) strcat(fstr,"s");
	
	result = simple_command(c,stoptime,"open %s %s %d\n",path,fstr,mode);
	if(result>=0) {
		if(get_stat_result(c,buf,stoptime)>=0) {
			return result;
		} else {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	} else {
		return result;
	}
}

int chirp_client_close( struct chirp_client *c, int fd, time_t stoptime  )
{
	return simple_command(c,stoptime,"close %d\n",fd);
}

int chirp_client_pread( struct chirp_client *c, int fd, void *buffer, int length, int offset, time_t stoptime  )
{
	int result;
	int actual;

	result = simple_command(c,stoptime,"pread %d %d %d\n",fd,length,offset);

	if( result>0 ) {
		actual = link_read(c->link,buffer,result,stoptime);
		if(actual!=result) {
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

int chirp_client_getfile( struct chirp_client *c, const char *path, FILE *stream, time_t stoptime  )
{
	int length;

	length = simple_command(c,stoptime,"getfile %s\n",path);

	if( length>0 ) {
		if(link_stream_to_file(c->link,stream,length,stoptime)==length) {
			return length;
		}
	}

	c->broken = 1;
	errno = ECONNRESET;
	return -1;
}

int chirp_client_readlink( struct chirp_client *c, const char *path, char *buffer, int length, time_t stoptime  )
{
	int result;
	int actual;

	result = simple_command(c,stoptime,"readlink %s %d\n",path,length);

	if( result>0 ) {
		actual = link_read(c->link,buffer,result,stoptime);
		if(actual!=result) {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
	}

	return result;
}

int chirp_client_canonicalize( struct chirp_client *c, const char *path, char *newpath, int length, time_t stoptime  )
{
	int result;
	int actual;

	result = simple_command(c,stoptime,"canonicalize %s %d\n",path,length);

	if( result>0 ) {
		actual = link_read(c->link,newpath,result,stoptime);
		if(actual!=result) {
			c->broken = 1;
			errno = ECONNRESET;
			return -1;
		}
		newpath[result] = 0;
	}

	return result;
}


int chirp_client_pwrite( struct chirp_client *c, int fd, const void *buffer, int length, int offset, time_t stoptime  )
{
	char line[CHIRP_LINE_MAX];
	int result;

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	sprintf(line,"pwrite %d %d %d\n",fd,length,offset);

	result = link_write(c->link,line,strlen(line),stoptime);
	if(result<0) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	result = link_write(c->link,buffer,length,stoptime);
	if(result!=length) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	return convert_result(get_result(c,stoptime));
}

int chirp_client_putfile( struct chirp_client *c, const char *path, FILE *stream, int mode, int length, time_t stoptime )
{

	char line[CHIRP_LINE_MAX];
	int result;

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	sprintf(line,"putfile %s %d %d\n",path,mode,length);

	result = link_write(c->link,line,strlen(line),stoptime);
	if(result<0) goto failure;

	result = get_result(c,stoptime);
	if(result<0) return convert_result(result);

	result = link_stream_from_file(c->link,stream,length,stoptime);
	if(result!=length) goto failure;
	
	return convert_result(get_result(c,stoptime));

	failure:
	c->broken = 1;
	errno = ECONNRESET;
	return -1;
}

int chirp_client_fchmod( struct chirp_client *c, int fd, int mode, time_t stoptime )
{
	return simple_command(c,stoptime,"fchmod %d %d\n",fd,mode);
}

int chirp_client_fchown( struct chirp_client *c, int fd, uid_t uid, gid_t gid, time_t stoptime )
{
	return simple_command(c,stoptime,"fchown %d %d %d\n",fd,uid,gid);
}

int chirp_client_ftruncate( struct chirp_client *c, int fd, int length, time_t stoptime )
{
	return simple_command(c,stoptime,"ftruncate %d %d\n",fd,length);
}

int chirp_client_fstat( struct chirp_client *c, int fd, struct stat *buf, time_t stoptime )
{
	int result = simple_command(c,stoptime,"fstat %d\n",fd);
	if(result>=0) result = get_stat_result(c,buf,stoptime);
	return result;
}

int chirp_client_stat( struct chirp_client *c, const char *path, struct stat *buf, time_t stoptime )
{
	int result = simple_command(c,stoptime,"stat %s\n",path);
	if(result>=0) result = get_stat_result(c,buf,stoptime);
	return result;
}

int chirp_client_lstat( struct chirp_client *c, const char *path, struct stat *buf, time_t stoptime )
{
	int result = simple_command(c,stoptime,"lstat %s\n",path);
	if(result>=0) result = get_stat_result(c,buf,stoptime);
	return result;
}

int chirp_client_fstatfs( struct chirp_client *c, int fd, struct statfs *buf, time_t stoptime )
{
	int result = simple_command(c,stoptime,"fstatfs %d\n",fd);
	if(result>=0) result = get_statfs_result(c,buf,stoptime);
	return result;
}

int chirp_client_statfs( struct chirp_client *c, const char *path, struct statfs *buf, time_t stoptime )
{
	int result = simple_command(c,stoptime,"statfs %s\n",path);
	if(result>=0) result = get_statfs_result(c,buf,stoptime);
	return result;
}

int chirp_client_unlink( struct chirp_client *c, const char *path, time_t stoptime )
{
	return simple_command(c,stoptime,"unlink %s\n",path);
}

int chirp_client_rename( struct chirp_client *c, const char *oldpath, const char *newpath, time_t stoptime )
{
	return simple_command(c,stoptime,"rename %s %s\n",oldpath,newpath);
}

int chirp_client_link( struct chirp_client *c, const char *oldpath, const char *newpath, time_t stoptime )
{
	return simple_command(c,stoptime,"link %s %s\n",oldpath,newpath);
}

int chirp_client_symlink( struct chirp_client *c, const char *oldpath, const char *newpath, time_t stoptime )
{
	return simple_command(c,stoptime,"symlink %s %s\n",oldpath,newpath);
}

int chirp_client_fsync( struct chirp_client *c, int fd, time_t stoptime )
{
	return simple_command(c,stoptime,"fsync %d\n",fd);
}

int chirp_client_mkdir( struct chirp_client *c, char const *path, int mode, time_t stoptime )
{
	return simple_command(c,stoptime,"mkdir %s %d\n",path,mode);
}

int chirp_client_rmdir( struct chirp_client *c, char const *path, time_t stoptime )
{
	return simple_command(c,stoptime,"rmdir %s\n",path);
}

int chirp_client_truncate( struct chirp_client *c, char const *path, int length, time_t stoptime )
{
	return simple_command(c,stoptime,"truncate %s %d\n",path,length);
}

int chirp_client_utime( struct chirp_client *c, char const *path, time_t actime, time_t modtime, time_t stoptime )
{
	return simple_command(c,stoptime,"utime %s %u %u\n",path,actime,modtime);
}

int chirp_client_access( struct chirp_client *c, char const *path, int mode, time_t stoptime )
{
	return simple_command(c,stoptime,"access %s %d\n",path,mode);
}

int chirp_client_chmod( struct chirp_client *c, char const *path, int mode, time_t stoptime )
{
	return simple_command(c,stoptime,"chmod %s %d\n",path,mode);
}

int chirp_client_chown( struct chirp_client *c, char const *path, uid_t uid, gid_t gid, time_t stoptime)
{
	return simple_command(c,stoptime,"chown %s %d %d\n",path,uid,gid);
}

int chirp_client_lchown( struct chirp_client *c, char const *path, uid_t uid, gid_t gid, time_t stoptime )
{
	return simple_command(c,stoptime,"lchown %s %d %d\n",path,uid,gid);
}

static int convert_result( int result )
{
	if(result>=0) {
		return result;
	} else {
		switch(result) {
			case CHIRP_ERROR_NOT_AUTHENTICATED:
			case CHIRP_ERROR_NOT_AUTHORIZED:
				errno = EACCES;
				break;
			case CHIRP_ERROR_DOESNT_EXIST:
				errno = ENOENT;
				break;
			case CHIRP_ERROR_ALREADY_EXISTS:
				errno = EEXIST;
				break;
			case CHIRP_ERROR_TOO_BIG:
				errno = EFBIG;
				break;
			case CHIRP_ERROR_NO_SPACE:
				errno = ENOSPC;
				break;
			case CHIRP_ERROR_NO_MEMORY:
				errno = ENOMEM;
				break;
			case CHIRP_ERROR_INVALID_REQUEST:
				errno = EINVAL;
				break;
			case CHIRP_ERROR_TOO_MANY_OPEN:
				errno = EMFILE;
				break;
			case CHIRP_ERROR_BUSY:
				errno = EBUSY;
				break;
			case CHIRP_ERROR_TRY_AGAIN:
				errno = EAGAIN;
				break;
			case CHIRP_ERROR_NOT_DIR:
				errno = ENOTDIR;
				break;
			case CHIRP_ERROR_IS_DIR:
				errno = EISDIR;
				break;
			case CHIRP_ERROR_NOT_EMPTY:
				errno = ENOTEMPTY;
				break;
			case CHIRP_ERROR_CROSS_DEVICE_LINK:
				errno = EXDEV;
				break;
			case CHIRP_ERROR_TIMED_OUT:
			case CHIRP_ERROR_DISCONNECTED:
			case CHIRP_ERROR_UNKNOWN:
				errno = ECONNRESET;
				break;
		}
		return -1;
	}
}

static int get_stat_result( struct chirp_client *c, struct stat *buf, time_t stoptime )
{
	char line[CHIRP_LINE_MAX];
	int fields;

	long long dev,ino,mode,nlink,uid,gid,rdev,size,blksize,blocks,atime,mtime,ctime;

	memset(buf,0,sizeof(*buf));

	if(!link_readline(c->link,line,CHIRP_LINE_MAX,stoptime)) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	fields = sscanf(line,"%lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld %lld\n",&dev,&ino,&mode,&nlink,&uid,&gid,&rdev,&size,&blksize,&blocks,&atime,&mtime,&ctime);
	if(fields!=13) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	buf->st_dev = dev;
	buf->st_ino = ino;
	buf->st_mode = mode;
	buf->st_nlink = nlink;
	buf->st_uid = uid;
	buf->st_gid = gid;
	buf->st_rdev = rdev;
	buf->st_size = size;
	buf->st_blksize = blksize;
	buf->st_blocks = blocks;
	buf->st_atime = atime;
	buf->st_mtime = mtime;
	buf->st_ctime = ctime;

	return 0;
}

static int get_statfs_result( struct chirp_client *c, struct statfs *buf, time_t stoptime )
{
	char line[CHIRP_LINE_MAX];
	int fields;
	long long type, bsize, blocks, bfree, bavail, files, ffree;

	memset(buf,0,sizeof(*buf));

	if(!link_readline(c->link,line,CHIRP_LINE_MAX,stoptime)) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	fields = sscanf(line,"%lld %lld %lld %lld %lld %lld %lld\n",&type,&bsize,&blocks,&bfree,&bavail,&files,&ffree);
	if(fields!=7) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}

	buf->f_type = type;
	buf->f_bsize = bsize;
	buf->f_bfree = bfree;
	buf->f_bavail = bavail;
	buf->f_files = files;
	buf->f_ffree = ffree;

	return 0;
}

static int get_result( struct chirp_client *c, time_t stoptime )
{
	char line[CHIRP_LINE_MAX];
	int result;
	int fields;

	if(!link_readline(c->link,line,CHIRP_LINE_MAX,stoptime)) {
		c->broken = 1;
		return CHIRP_ERROR_DISCONNECTED;
	}

	fields = sscanf(line,"%d",&result);
	if(fields!=1) {
		c->broken = 1;
		return CHIRP_ERROR_DISCONNECTED;
	}

	return result;
}

/*
  vsprintf_chirp -- simple sprintf capabilities with character escaping

  The following format characters are interpreted:

  %d -- decimal
  %s -- word (whitespace is escaped)
  %% -- output %
*/

void vsprintf_chirp(char *command,char const *fmt,va_list args)
{
	char       *c;
	char const *f;

	c = command;
	f = fmt;
	while(*f) {
		if(*f == '%') {
			switch(*(++f)) {
			case 'd':
				f++;
				sprintf(c,"%d",va_arg(args,int));
				c += strlen(c);
				break;
			case 'u':
				f++;
				sprintf(c,"%u",va_arg(args,int));
				c += strlen(c);
				break;
			case 's': {
				char const *w = va_arg(args,char const *);
				f++;
				while(*w) {
					switch(*w) {
					case ' ':
					case '\t':
					case '\n':
					case '\r':
					case '\\':
						*(c++) = '\\';
						/*fall through*/
					default:
						*(c++) = *(w++);
					}
				}
				break;
			}
			case '%':
				*(c++) = *(f++);
				break;
			default:
				abort();
			}
		} else {
			*(c++) = *(f++);
		}
	}
	*(c++) = '\0';
}

int simple_command( struct chirp_client *c, time_t stoptime, char const *fmt, ... )
{
	int     result;
	char    command[CHIRP_LINE_MAX];
	va_list args;

	va_start(args,fmt);
	vsprintf_chirp(command,fmt,args);
	va_end(args);

	if(c->broken) {
		errno = ECONNRESET;
		return -1;
	}

	result = link_write(c->link,command,strlen(command),stoptime);
	if(result < 0) {
		c->broken = 1;
		errno = ECONNRESET;
		return -1;
	}
	return convert_result(get_result(c,stoptime));
}
