/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "chirp_reli.h"
#include "chirp_protocol.h"
#include "chirp_client.h"

#include "macros.h"
#include "debug.h"
#include "full_io.h"
#include "sleeptools.h"
#include "hash_table.h"
#include "xmalloc.h"

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>

#define MIN_DELAY 1
#define MAX_DELAY 60
#define MINOR_TIMEOUT MIN(time(0)+60,stoptime)

struct chirp_file {
	char host[CHIRP_LINE_MAX];
	char path[CHIRP_LINE_MAX];
	struct stat info;
	int fd;
	int flags;
	int mode;
	int serial;
	int stale;
};

struct hash_table *table = 0;

static struct chirp_client * connect_to_host( const char *host, time_t stoptime )
{
	struct chirp_client *c;

	if(!table) {
		table = hash_table_create(0,0);
		if(!table) return 0;
	}

	c = hash_table_lookup(table,host);
	if(c) return c;
	
	if(!strcmp(host,"CONDOR")) {
		c = chirp_client_connect_condor(stoptime);
	} else {
		c = chirp_client_connect(host,1,stoptime);
	}

	if(c) {
		hash_table_insert(table,host,c,(void**)0);
		return c;
	} else {
		return 0;
	}
}

static int connect_to_file( struct chirp_client *client, struct chirp_file *file, time_t stoptime )
{
	struct stat buf;

	if(file->stale) {
		errno = ESTALE;
		return -1;
	}

	if(chirp_client_serial(client)==file->serial) return 1;

	debug(D_CHIRP,"verifying: %s",file->path);
	file->fd = chirp_client_open(client,file->path,file->flags,file->mode,&buf,stoptime);
	file->serial = chirp_client_serial(client);
	if(file->fd>=0) {
		if(buf.st_dev!=file->info.st_dev) {
			debug(D_CHIRP,"stale: device changed: %s",file->path);
			file->stale = 1;
			errno = ESTALE;
			return 0;
		} else if(buf.st_ino!=file->info.st_ino) {
			debug(D_CHIRP,"stale: inode changed: %s",file->path);
			file->stale = 1;
			errno = ESTALE;
			return 0;
		} else if(buf.st_rdev!=file->info.st_rdev) {
			debug(D_CHIRP,"stale: rdev changed: %s",file->path);
			file->stale = 1;
			errno = ESTALE;
			return 0;
		} else {
			debug(D_CHIRP,"uptodate: %s",file->path);
			file->stale = 0;
			return 1;
		}
	} else {
		if(errno!=ECONNRESET) {
			debug(D_CHIRP,"stale: %s: %s",strerror(errno),file->path);
			file->stale = 1;
			errno = ESTALE;
			return 0;
		}
	}

	return 1;
}

static void invalidate_host( const char *host )
{
	struct chirp_client *c;
	c = hash_table_remove(table,host);
	if(c) chirp_client_disconnect(c);
}

struct chirp_file * chirp_reli_open( const char *host, const char *path, int flags, int mode, time_t stoptime )
{
	struct chirp_file *file;
	int delay=MIN_DELAY;
	int nexttry;
	int result;
	struct stat buf;
	time_t current;

	while(1) {
		struct chirp_client *client = connect_to_host(host,MINOR_TIMEOUT);
		if(client) {
			result = chirp_client_open(client,path,flags,mode,&buf,MINOR_TIMEOUT);
			if(result>=0) {
				file = xmalloc(sizeof(*file));
				strcpy(file->host,host);
				strcpy(file->path,path);
				memcpy(&file->info,&buf,sizeof(buf));
				file->fd = result;
				file->flags = flags & ~(O_CREAT|O_TRUNC|O_EXCL);
				file->mode = mode;
				file->serial = chirp_client_serial(client);
				file->stale = 0;
				return file;
			} else {
				if(errno!=ECONNRESET) return 0;
			}
	 		invalidate_host(host);
		}
		if(time(0)>=stoptime) {
			errno = ECONNRESET;
			return 0;
		}
		debug(D_CHIRP,"couldn't talk to %s: %s\n",host,strerror(errno));
		current = time(0);
		nexttry = MIN(stoptime,current+delay);
		debug(D_CHIRP,"try again in %d seconds\n",nexttry-current);
		sleep_until(nexttry);
		delay = MIN(delay*2,MAX_DELAY);
	}
}

int chirp_reli_close( struct chirp_file *file, time_t stoptime )
{
	struct chirp_client *client = connect_to_host(file->host,MINOR_TIMEOUT);
	if(client) {
		if(chirp_client_serial(client)==file->serial) {
			chirp_client_close(client,file->fd,MINOR_TIMEOUT);
		}
	}
	free(file);
	return 0;
}

#define RETRY_FILE( ZZZ ) \
	int delay=MIN_DELAY; \
	int nexttry; \
	int result; \
	time_t current; \
	while(1) { \
		struct chirp_client *client = connect_to_host(file->host,MINOR_TIMEOUT); \
		if(client) { \
			if(connect_to_file(client,file,MINOR_TIMEOUT)) { \
				ZZZ \
				if(result>=0 || errno!=ECONNRESET) return result; \
			} \
			if(errno==ESTALE) return -1; \
	 		invalidate_host(file->host); \
		} else { \
			if(errno==ENOENT) return -1; \
			if(errno==EPERM) return -1; \
		} \
		if(time(0)>=stoptime) { \
			errno = ECONNRESET; \
			return -1; \
		} \
		debug(D_CHIRP,"couldn't talk to %s: %s\n",file->host,strerror(errno)); \
		current = time(0); \
		nexttry = MIN(stoptime,current+delay); \
		debug(D_CHIRP,"try again in %d seconds\n",nexttry-current); \
		sleep_until(nexttry); \
		delay = MIN(delay*2,MAX_DELAY); \
	}


int chirp_reli_pread( struct chirp_file *file, void *buffer, int length, int offset, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_pread(client,file->fd,buffer,length,offset,MINOR_TIMEOUT); )
}

int chirp_reli_pwrite( struct chirp_file *file, const void *buffer, int length, int offset, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_pwrite(client,file->fd,buffer,length,offset,MINOR_TIMEOUT); )
}

int chirp_reli_fstat( struct chirp_file *file, struct stat *buf, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_fstat(client,file->fd,buf,MINOR_TIMEOUT); )
}

int chirp_reli_fstatfs( struct chirp_file *file, struct statfs *buf, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_fstatfs(client,file->fd,buf,MINOR_TIMEOUT); )
}

int chirp_reli_fchown( struct chirp_file *file, uid_t uid, gid_t gid, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_fchown(client,file->fd,uid,gid,MINOR_TIMEOUT); )
}

int chirp_reli_fchmod( struct chirp_file *file, int mode, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_fchmod(client,file->fd,mode,MINOR_TIMEOUT); )
}

int chirp_reli_ftruncate( struct chirp_file *file, int length, time_t stoptime )
{
	RETRY_FILE( result = chirp_client_ftruncate(client,file->fd,length,MINOR_TIMEOUT); )
}

#define RETRY_ATOMIC( ZZZ ) \
	int delay=MIN_DELAY; \
	int nexttry; \
	int result; \
	time_t current; \
	while(1) { \
		struct chirp_client *client = connect_to_host(host,MINOR_TIMEOUT); \
		if(client) { \
			ZZZ \
			if(result>=0 || errno!=ECONNRESET) return result; \
 			invalidate_host(host); \
		} else { \
			if(errno==ENOENT) return -1; \
			if(errno==EPERM) return -1; \
		} \
		if(time(0)>=stoptime) { \
			errno = ECONNRESET; \
			return -1; \
		} \
		debug(D_CHIRP,"couldn't talk to %s: %s\n",host,strerror(errno)); \
		current = time(0); \
		nexttry = MIN(stoptime,current+delay); \
		debug(D_CHIRP,"try again in %d seconds\n",nexttry-current); \
		sleep_until(nexttry); \
		delay = MIN(delay*2,MAX_DELAY); \
	}

int chirp_reli_lookup( const char *host, const char *path, char **redirect, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_lookup(client,path,redirect,MINOR_TIMEOUT); );
}

int chirp_reli_getfile( const char *host, const char *path, FILE *stream, time_t stoptime )
{
	RETRY_ATOMIC( fseek(stream,0,SEEK_SET); result = chirp_client_getfile(client,path,stream,MINOR_TIMEOUT); )
}


int chirp_reli_putfile( const char *host, const char *path, FILE *stream, int mode, int length, time_t stoptime )
{
	RETRY_ATOMIC( fseek(stream,0,SEEK_SET); result = chirp_client_putfile(client,path,stream,mode,length,MINOR_TIMEOUT); )
}

int chirp_reli_getdir( const char *host, const char *path, chirp_dir_t callback, void *arg, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_getdir(client,path,callback,arg,MINOR_TIMEOUT); )
}

int chirp_reli_unlink( const char *host, const char *path, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_unlink(client,path,MINOR_TIMEOUT); )
}

int chirp_reli_rename( const char *host, const char *path, const char *newpath, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_rename(client,path,newpath,MINOR_TIMEOUT); )
}

int chirp_reli_link( const char *host, const char *path, const char *newpath, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_link(client,path,newpath,MINOR_TIMEOUT); )
}

int chirp_reli_symlink( const char *host, const char *path, const char *newpath, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_symlink(client,path,newpath,MINOR_TIMEOUT); )
}

int chirp_reli_readlink( const char *host, const char *path, char *buf, int length, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_readlink(client,path,buf,length,MINOR_TIMEOUT); )
}

int chirp_reli_mkdir( const char *host, char const *path, int mode, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_mkdir(client,path,mode,MINOR_TIMEOUT); )
}

int chirp_reli_rmdir( const char *host, char const *path, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_rmdir(client,path,MINOR_TIMEOUT); )
}

int chirp_reli_stat( const char *host, const char *path, struct stat *buf, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_stat(client,path,buf,MINOR_TIMEOUT); )
}

int chirp_reli_lstat( const char *host, const char *path, struct stat *buf, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_lstat(client,path,buf,MINOR_TIMEOUT); )
}

int chirp_reli_statfs( const char *host, const char *path, struct statfs *buf, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_statfs(client,path,buf,MINOR_TIMEOUT); )
}

int chirp_reli_access( const char *host, const char *path, int mode, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_access(client,path,mode,MINOR_TIMEOUT); )
}

int chirp_reli_chmod( const char *host, const char *path, int mode, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_chmod(client,path,mode,MINOR_TIMEOUT); )
}

int chirp_reli_chown( const char *host, const char *path, uid_t uid, gid_t gid, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_chown(client,path,uid,gid,MINOR_TIMEOUT); )
}

int chirp_reli_lchown( const char *host, const char *path, uid_t uid, gid_t gid, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_lchown(client,path,uid,gid,MINOR_TIMEOUT); )
}

int chirp_reli_truncate( const char *host, const char *path, int length, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_truncate(client,path,length,MINOR_TIMEOUT); )
}

int chirp_reli_utime( const char *host, const char *path, time_t actime, time_t modtime, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_utime(client,path,actime,modtime,MINOR_TIMEOUT); )
}

int chirp_reli_canonicalize( const char *host, const char *path, char *newpath, int length, time_t stoptime )
{
	RETRY_ATOMIC( result = chirp_client_canonicalize(client,path,newpath,length,MINOR_TIMEOUT); )
}


