/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "chirp_client.h"
#include "chirp_protocol.h"
#include "timestamp.h"
#include "debug.h"
#include "auth_all.h"

#ifdef HAS_LIBREADLINE
#include "readline/readline.h"
#include "readline/history.h"
#endif

#define BUFFER_SIZE 32768

static int do_open( const char *x, const char *y );
static int do_close( const char *x, const char *y );
static int do_get( const char *x, const char *y );
static int do_put( const char *x, const char *y );
static int do_cd( const char *x, const char *y );
static int do_lcd( const char *x, const char *y );
static int do_pwd( const char *x, const char *y );
static int do_lpwd( const char *x, const char *y );
static int do_ls( const char *x, const char *y );
static int do_mv( const char *x, const char *y );
static int do_rm( const char *x, const char *y );
static int do_mkdir( const char *x, const char *y );
static int do_rmdir( const char *x, const char *y );
static int do_stat( const char *x, const char *y );
static int do_timeout( const char *x, const char *y );
static int do_debug( const char *x, const char *y );
static int do_help( const char *x, const char *y );
static int do_quit( const char *x, const char *y );

static struct chirp_client *client=0;
static int timeout=60;
static char current_host[CHIRP_PATH_MAX];
static char current_local_dir[CHIRP_PATH_MAX];
static char current_remote_dir[CHIRP_PATH_MAX];

struct command {
	char *name;
	int must_be_open;
	int minargs, maxargs;
	char *help;
	int (*handler) ( const char *arg1, const char *arg2 );
};

static struct command list[] =
{
{"open",    0, 1, 1, "<host>",                   do_open},
{"close",   1, 0, 0, "",                         do_close},
{"cd",      1, 1, 1, "<remotedir>",              do_cd},
{"lcd",     0, 1, 1, "<localdir>",               do_lcd},
{"pwd",     1, 0, 0, "",                         do_pwd},
{"lpwd",    0, 0, 0, "",                         do_lpwd},
{"get",	    1, 1, 2, "<remotefile> [localfile]", do_get},
{"put",	    1, 1, 2, "<localfile> [remotefile]", do_put},
{"ls",      1, 0, 1, "[remotepath]",             do_ls},
{"mv",      1, 2, 2, "<oldname> <newname>",      do_mv},
{"rm",	    1, 1, 1, "<file>",                   do_rm},
{"mkdir",   1, 1, 1, "<dir>",                    do_mkdir},
{"rmdir",   1, 1, 1, "<dir>",                    do_rmdir},
{"stat",    1, 1, 1, "<file>",                   do_stat},
{"timeout", 0, 1, 1, "<seconds>",                do_timeout},
{"debug",   0, 0, 1, "[subsystem]",              do_debug},
{"help",    0, 0, 0, "",                         do_help},
{"quit",    0, 0, 0, "",                         do_quit},
{0,0,0,0,0},
};

int main( int argc, char *argv[] )
{	
#ifdef HAS_LIBREADLINE
	char *temp;
#endif
	int interactive_mode;
	char prompt[CHIRP_LINE_MAX];
	char line[CHIRP_LINE_MAX];
	char command[CHIRP_LINE_MAX];
	char arg1[CHIRP_LINE_MAX];
	char arg2[CHIRP_LINE_MAX];
	int fields;
	int i;

	debug_config(argv[0]);
	auth_register_all();

	interactive_mode = isatty(0);
	getcwd(current_local_dir,CHIRP_PATH_MAX);

	if(argc>1) {
		if(do_open(argv[1],0)<0) {
			fprintf(stderr,"couldn't open %s: %s\n",argv[1],strerror(errno));
			return 1;
		}
	}

	while(1) {
		if(interactive_mode) {
			sprintf(prompt," chirp:%s:%s> ",current_host,current_remote_dir);
		} else {
			prompt[0] = 0;
		}

#ifdef HAS_LIBREADLINE
		temp = readline(prompt);		
		if(!temp) break;
		strcpy(line,temp);
		free(temp);
#else
		printf("%s",prompt);
		fflush(stdout);
		if(!fgets(line,CHIRP_LINE_MAX,stdin)) break;
#endif

		if(!line[0]) continue;

		add_history(line);		
		command[0] = 0;
		arg1[0] = 0;
		arg2[0] = 0;

		fields = sscanf(line,"%s %s %s",command,arg1,arg2);
		if(fields==0) continue;

		for(i=0;list[i].name;i++) {
			if(!strcmp(list[i].name,command)) {
				if(!client && list[i].must_be_open) {
					printf("not connected\n");
					break;
				} else if( ((fields-1)>=list[i].minargs) && ((fields-1)<=list[i].maxargs ) ) {
					if(list[i].handler(arg1,arg2)<0) {
						fprintf(stderr,"couldn't %s %s %s: %s\n",command,arg1,arg2,strerror(errno));
						if(!interactive_mode) return 1;
					}
					break;
				} else {
					printf("use: %s %s\n",command,list[i].help);
					break;
				}
			}
		}

		if(!list[i].name) {
			printf("unknown command: %s  (try 'help')\n",command);
		}
	}

	printf("\n");
	return 0;
}

static int do_open( const char *host, const char *x )
{
	client = chirp_client_connect(host,1,time(0)+timeout);
	if(!client) {
		return -1;
	} else {
		printf("connected to %s\n",host);
		strcpy(current_host,host);
		current_remote_dir[0] = '/';
		return 0;
	}
}

static int do_close( const char *x, const char *y )
{
	if(client) {
		chirp_client_disconnect(client);
		current_host[0] = 0;
		client = 0;
	}
	return 0;
}

static void complete_local_path( const char *file, char *full_path )
{
	if(file[0]!='/') {
		sprintf(full_path,"%s/%s",current_local_dir,file);
	} else {
		strcpy(full_path,file);
	}
}

static void complete_remote_path( const char *file, char *full_path )
{
	if(file[0]!='/') {
		sprintf(full_path,"%s/%s",current_remote_dir,file);
	} else {
		strcpy(full_path,file);
	}
}

static int do_cd( const char *file, const char *x )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	if(chirp_client_canonicalize(client,full_path,current_remote_dir,CHIRP_PATH_MAX,time(0)+timeout)<0) {
		return -1;
	} else {
		return 0;
	}
}

static int do_lcd( const char *file, const char *x )
{
	char full_path[CHIRP_PATH_MAX];
	complete_local_path(file,full_path);
	if(chdir(full_path)==0) {
		getcwd(current_local_dir,CHIRP_PATH_MAX);
		return 0;
	} else {
		return -1;
	}
}

static int do_pwd( const char *x, const char *y )
{
	printf("%s\n",current_remote_dir);
	return 0;
}

static int do_lpwd( const char *x, const char *y )
{
	printf("%s\n",current_local_dir);
	return 0;
}

static int do_get( const char *remote_file, const char *lpath )
{
	FILE *file;
	timestamp_t start;
	double elapsed;
	int length;
	int save_errno;

	char remote_full_path[CHIRP_PATH_MAX];
	char local_full_path[CHIRP_PATH_MAX];

	if(!lpath[0]) lpath=remote_file;

	complete_local_path(lpath,local_full_path);
	complete_remote_path(remote_file,remote_full_path);

	file = fopen(local_full_path,"wb+");
	if(!file) return -1;

	start = timestamp_get();

	length = chirp_client_getfile(client,remote_full_path,file,time(0)+timeout);
	if(length<0) {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}

	elapsed = (timestamp_get()-start)/1000000.0;

	if(length>=0) {
		printf("got %d bytes in %.2lf seconds (%.2lf MB/s)\n",length,elapsed,length/elapsed/1000000.0);
		fclose(file);
		return 0;
	} else {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}
}

static int do_put( const char *local_file, const char *remote_file )
{
	FILE *file;
	timestamp_t start;
	double elapsed;
	int length;
	int save_errno;

	char remote_full_path[CHIRP_PATH_MAX];
	char local_full_path[CHIRP_PATH_MAX];

	if(!local_file[0]) local_file=remote_file;

	complete_local_path(local_file,local_full_path);
	complete_remote_path(remote_file,remote_full_path);

	file = fopen(local_full_path,"r");
	if(!file) return -1;

	fseek(file,0,SEEK_END);
	length = ftell(file);
	fseek(file,0,SEEK_SET);

	start = timestamp_get();

	length = chirp_client_putfile(client,remote_full_path,file,0777,length,time(0)+timeout);
	if(length<0) {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}

	elapsed = (timestamp_get()-start)/1000000.0;

	if(length>=0) {
		printf("put %d bytes in %.2lf seconds (%.2lf MB/s)\n",length,elapsed,length/elapsed/1000000.0);
		fclose(file);
		return 0;
	} else {
		save_errno = errno;
		fclose(file);
		errno = save_errno;
		return -1;
	}
}

static void do_ls_item( const char *path, void *arg )
{
	printf(" %s\n",path);
}

static int do_ls( const char *file, const char *y )
{
	char full_path[CHIRP_PATH_MAX];
	if(!file[0]) file = ".";
	complete_remote_path(file,full_path);
	return chirp_client_getdir(client,full_path,do_ls_item,0,time(0)+timeout);
}

static int do_rm( const char *file, const char *x )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	return chirp_client_unlink(client,full_path,time(0)+timeout);
}

static int do_rmdir( const char *file, const char *x )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	return chirp_client_rmdir(client,full_path,time(0)+timeout);
}

static int do_mkdir( const char *file, const char *x )
{
	char full_path[CHIRP_PATH_MAX];
	complete_remote_path(file,full_path);
	return chirp_client_mkdir(client,full_path,0777,time(0)+timeout);
}

static int do_stat( const char *file, const char *x )
{
	char full_path[CHIRP_PATH_MAX];
	struct stat info;

	complete_remote_path(file,full_path);

	if(chirp_client_stat(client,full_path,&info,time(0)+timeout)<0) {
		return -1;
	} else {

		/*
		Note that we deliberately cast call integers into long long
		because these types vary widely from platform to platform and
		may be 64 bit in the future.  This is overkill but foolproof.
		*/

		printf("device:  %lld\n",(long long)info.st_dev);
		printf("inode:   %lld\n",(long long)info.st_ino);
		printf("mode:    %04llo\n",(long long)info.st_mode);
		printf("nlink:   %lld\n",(long long)info.st_nlink);
		printf("uid:     %lld\n",(long long)info.st_uid);
		printf("gid:     %lld\n",(long long)info.st_gid);
		printf("rdevice: %lld\n",(long long)info.st_rdev);
		printf("size:    %lld\n",(long long)info.st_size);
		printf("blksize: %lld\n",(long long)info.st_blksize);
		printf("blocks:  %lld\n",(long long)info.st_blocks);
		printf("atime:   %s",ctime(&(info.st_atime)));
		printf("mtime:   %s",ctime(&(info.st_mtime)));
		printf("ctime:   %s",ctime(&(info.st_ctime)));

		return 0;
	}
 }

static int do_mv( const char *old_file, const char *new_file )
{
	char old_full_path[CHIRP_PATH_MAX];
	char new_full_path[CHIRP_PATH_MAX];
	complete_remote_path(old_file,old_full_path);
	complete_remote_path(new_file,new_full_path);
	return chirp_client_rename(client,old_full_path,new_full_path,time(0)+timeout);
}

static int do_debug( const char *flag, const char *y )
{
	if(flag[0]) {
		if(debug_flags_set(flag)) {
			return 0;
		} else {
			printf("Valid debugging flags are: ");
			debug_flags_print(stdout);
			printf("\n");
			errno = EINVAL;
			return -1;
		}
	} else {
		debug_flags_clear();
		return 0;
	}	
	
}

static int do_timeout( const char *x, const char *y )
{
	timeout = atoi(x);
	printf("Timeout is %d seconds.\n",timeout);
	return 0;
}

static int do_quit( const char *x, const char *y )
{
	exit(0);
	return -1;
}

static int do_help( const char *x, const char *y )
{
	int i;
	printf("Commands:\n");
	for(i=0;list[i].name;i++) {
		printf("  %8s %s\n",list[i].name,list[i].help);
	}
	printf("\nDebugging subsystems are:\n");
	debug_flags_print(stdout);
	printf("\n");
	return 0;
}

