/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/


#include "auth.h"
#include "debug.h"
#include "stringtools.h"
#include "domain_name_cache.h"
#include "xmalloc.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

struct auth_ops {
	char *type;
	auth_assert_t assert;
	auth_accept_t accept;
	struct auth_ops *next;
};

static struct auth_ops *list=0;

static struct auth_ops * type_lookup( char *type )
{
	struct auth_ops *a;

	for(a=list;a;a=a->next) {
		if(!strcmp(a->type,type)) return a;
	}

	return 0;
}

int auth_lookup( const char *authfile, const char *type, const char *host, const char *subject, char **local_user )
{
	char line[AUTH_LINE_MAX];
	char ntype[AUTH_LINE_MAX];
	char nhost[AUTH_LINE_MAX];
	char nsubject[AUTH_LINE_MAX];
	char lsubject[AUTH_LINE_MAX];
	int fields;
	int result=0;
	FILE *file;
	int linenum=0;

	debug(D_AUTH,"looking in %s to see if %s is authorized\n",authfile,subject);

	file = fopen(authfile,"r");
	if(!file) {
		debug(D_AUTH,"couldn't open %s: %s",authfile,strerror(errno));
		return 0;
	}

	while(fgets(line,sizeof(line),file)) {

		linenum++;
		string_chomp(line);

		if( line[0]=='#' ) continue;

		fields = sscanf(line,"%[^:]:%[^:]:%[^:]:%[^:]",ntype,nhost,nsubject,lsubject);
		if(fields==4) {
			if(!string_match(ntype,type)) {
				debug(D_AUTH,"line %d doesn't match: type %s!=%s'\n",linenum,ntype,type);
				continue;
			}
			if(!string_match(nhost,host)) {
				debug(D_AUTH,"line %d doesn't match: host %s!=%s\n",linenum,nhost,host);
				continue;
			}
			if(!string_match(nsubject,subject)) {
				debug(D_AUTH,"line %d doesn't match: subject %s!=%s",linenum,nsubject,subject);
				continue;
			}

			if(!strcmp(lsubject,"*")) {
				*local_user = xstrdup(subject);
			} else {
				*local_user = xstrdup(lsubject);
			}
			debug(D_AUTH,"line %d: %s is authorized as %s\n",linenum,subject,*local_user);
			result = 1;
			break;
		} else if(fields<=1) {
			// nothing
		} else {
			debug(D_AUTH,"auth_lookup: bad data on line %d in file %s",linenum,authfile);
		}
	}

	fclose(file);
	return result;
}

int auth_assert( struct link *link, char **type, char **subject, time_t stoptime )
{
	char line[AUTH_LINE_MAX];
	struct auth_ops *a;

	for( a=list; a; a=a->next ) {

		debug(D_AUTH,"requesting '%s' authentication",a->type);

		sprintf(line,"%s\n",a->type);
		link_write(link,line,strlen(line),stoptime);

		if(!link_readline(link,line,AUTH_LINE_MAX,stoptime)) break;
		if(!strcmp(line,"yes")) {
			debug(D_AUTH,"server agrees to try '%s'",a->type);
			if(a->assert(link,stoptime)) {
				debug(D_AUTH,"successfully authenticated");
				if(!link_readline(link,line,AUTH_LINE_MAX,stoptime)) break;
				if(!strcmp(line,"yes")) {
					debug(D_AUTH,"reading back auth info from server");
					if(!link_readline(link,line,sizeof(line),stoptime)) return 0;
					*type = xstrdup(line);
					if(!link_readline(link,line,sizeof(line),stoptime)) return 0;
					*subject = xstrdup(line);
					return 1;
				} else {
					debug(D_AUTH,"but not authorized to continue");
				}
			} else {
				debug(D_AUTH,"failed to authenticate");
			}
		} else {
			debug(D_AUTH,"server refuses to try '%s'",a->type);
		}
		debug(D_AUTH,"still trying...");
	}

	if(!a) {
		debug(D_AUTH,"ran out of authenticators");
	} else {
		debug(D_AUTH,"lost connection");
	}

	return 0;
}

int auth_accept( struct link *link, const char *authfile, char **typeout, char **hostout, char **subject, char **local_user, time_t stoptime )
{
	struct auth_ops *a;
	char type[AUTH_TYPE_MAX];
	char line[AUTH_LINE_MAX];
	char addr[LINK_ADDRESS_MAX];
	char name[DOMAIN_NAME_MAX];
	int port;

	*subject = 0;
	link_address_remote(link,addr,&port);
	if(domain_name_cache_lookup_reverse(addr,name)) {
		debug(D_AUTH,"client '%s' is '%s'",addr,name);
	} else {
		debug(D_AUTH,"client '%s' has no domain name",addr);
		strcpy(name,addr);
	}


	while(link_readline(link,type,AUTH_TYPE_MAX,stoptime)) {

		string_chomp(line);

		debug(D_AUTH,"%s requests '%s' authentication",name,type);

		a = type_lookup( type );
		if(a) {
			debug(D_AUTH,"I agree to try '%s' ",type);
			link_write(link,"yes\n",4,stoptime);
		} else {
			debug(D_AUTH,"I do not agree to '%s' ",type);
			link_write(link,"no\n",3,stoptime);
			continue;
		}

		if(a->accept(link,subject,stoptime)) {
			debug(D_AUTH,"'%s' authentication succeeded",type);
			debug(D_AUTH,"%s is '%s'",name,*subject);
			if(!authfile || auth_lookup(authfile,type,name,*subject,local_user)) {
				debug(D_AUTH,"%s is authorized to continue",name);
				sprintf(line,"yes\n");
				link_write(link,line,strlen(line),stoptime);
				sprintf(line,"%s\n",type);
				link_write(link,line,strlen(line),stoptime);
				sprintf(line,"%s\n",*subject);
				link_write(link,line,strlen(line),stoptime);
	       			*typeout = xstrdup(type);
	       			*hostout = xstrdup(name);
				return 1;
			} else {
				debug(D_AUTH,"but %s:%s:%s is not authorized to continue",type,name,*subject);
				free(*subject);
				sprintf(line,"no\n");
				link_write(link,line,strlen(line),stoptime);
			}
		} else {
			debug(D_AUTH,"%s could not authenticate using '%s'",name,type);
		}
		debug(D_AUTH,"still trying");
	}

	debug(D_AUTH,"%s disconnected",name);

	return 0;
}


int auth_barrier( struct link *link, const char *response, time_t stoptime )
{
	char line[AUTH_LINE_MAX];

	link_write(link,response,strlen(response),stoptime);

	if(link_readline(link,line,sizeof(line),stoptime)) {
		if(!strcmp(line,"yes")) {
			return 1;
		}
	}

	return 0;
}

int auth_register( char *type, auth_assert_t assert, auth_accept_t accept )
{
	struct auth_ops *a, *l;

	a = (struct auth_ops *) malloc(sizeof(struct auth_ops));
	if(!a) return 0;

	a->type = type;
	a->assert = assert;
	a->accept = accept;
	a->next = 0;

	if(!list) {
		list = a;
	} else {
		l = list;
		while(l->next) {
			l = l->next;
		}
		l->next = a;
	}

	return 1;
}

void auth_clear()
{
	struct auth_ops *n;

	while(list) {
		n = list->next;
		free(list);
		list = n;
	}
}
