/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "catalog_server.h"
#include "datagram.h"
#include "link.h"
#include "hash_cache.h"
#include "debug.h"
#include "getopt.h"
#include "nvpair.h"
#include "stringtools.h"
#include "domain_name_cache.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <alloca.h>
#include <errno.h>
#include <signal.h>
#include <unistd.h>

/* The table of nvpairs, hashed on address:port */
static struct hash_cache *table=0;

/* The time for which updated data lives before automatic deletion */
static int lifetime = 900;

/* The port upon which to listen. */
static int port = CATALOG_PORT;

/* This machine's canonical name. */
static char hostname[DOMAIN_NAME_MAX];

static void handle_updates( struct datagram *update_port )
{
	char data[DATAGRAM_PAYLOAD_MAX*2];
	char addr[DATAGRAM_ADDRESS_MAX];
	char key[DATAGRAM_ADDRESS_MAX*2];
	int port;
	int result;
	struct nvpair *nv;
	char *name;

	while(1) {
		result = datagram_recv(update_port,data,DATAGRAM_PAYLOAD_MAX,addr,&port,0);
		if(result<=0) return;

		data[result]=0;

		nv = nvpair_create();
		nvpair_parse(nv,data);
		nvpair_insert_string(nv,"address",addr);
		nvpair_insert_integer(nv,"lastheardfrom",time(0));

		name = nvpair_lookup_string(nv,"name");
		if(!name) name = "???";

		sprintf(key,"%s:%s",addr,name);
		hash_cache_insert(table,key,nv,lifetime);

		debug(D_DEBUG,"received udp update from %s",key);
	}
}

static int send_text( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_text(nv,stream);
	return 1;
}

static int send_xml( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_xml(nv,stream);
	return 1;
}

static int send_old_classads( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_old_classads(nv,stream);
	return 1;
}

static int send_new_classads( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_new_classads(nv,stream);
	return 1;
}

static const char *html_headers[] = {
	"type", "name", "owner", "total", "avail", "url", 0,
};

static int send_html( struct hash_cache *table, char *key, void *nv, void *stream )
{
	nvpair_print_html(nv,stream,html_headers);
	return 1;
}

static void handle_query( struct link *query_link )
{
	FILE *stream;
	char line[LINE_MAX];
	char action[LINE_MAX];
	char path[LINE_MAX];
	char version[LINE_MAX];
	char addr[LINK_ADDRESS_MAX];
	int port;

	link_address_remote(query_link,addr,&port);

	debug(D_DEBUG,"www query from %s:%d",addr,port);

	link_nonblocking(query_link,0);
	stream = fdopen(link_fd(query_link),"r+");
	if(!stream) return;

	if(!fgets(line,sizeof(line),stream)) return;
	string_chomp(line);
	if(sscanf(line,"%s %s %s",action,path,version)!=3) return;

	while(1) {
		if(!fgets(line,sizeof(line),stream)) return;
		if(line[0]=='\n' || line[0]=='\r') break;
	}

	fprintf(stream,"HTTP/0.9 200 OK\n");
	fprintf(stream,"Server: catalog_server\n");
	fprintf(stream,"Connection: close\n");
	fprintf(stream,"Content-type: text/");

	if(!strcmp(path,"/query.text")) {
		fprintf(stream,"plain\n\n");
		hash_cache_iterate(table,send_text,stream);
	} else if(!strcmp(path,"/query.oldclassads")) {
		fprintf(stream,"plain\n\n");
		hash_cache_iterate(table,send_old_classads,stream);
	} else if(!strcmp(path,"/query.newclassads")) {
		fprintf(stream,"plain\n\n");
		hash_cache_iterate(table,send_new_classads,stream);
	} else if(!strcmp(path,"/query.xml")) {
		fprintf(stream,"xml\n\n");
		fprintf(stream,"<?xml version=\"1.0\" standalone=\"yes\"?>\n");
		fprintf(stream,"<catalog>\n");
		hash_cache_iterate(table,send_xml,stream);
		fprintf(stream,"</catalog>\n");
	} else {
		fprintf(stream,"html\n\n");
		fprintf(stream,"<title>%s catalog server</title>\n",hostname);
		fprintf(stream,"<center>\n");
		fprintf(stream,"<h1>%s catalog server</h1>\n",hostname);
		fprintf(stream,"<a href=/query.text>text</a> - ");
		fprintf(stream,"<a href=/query.html>html</a> - ");
		fprintf(stream,"<a href=/query.xml>xml</a> - ");
		fprintf(stream,"<a href=/query.oldclassads>oldclassads</a> - ");		
		fprintf(stream,"<a href=/query.newclassads>newclassads</a>");		
		fprintf(stream,"<p>\n");

		nvpair_print_html_header(stream,html_headers);
		hash_cache_iterate(table,send_html,stream);
		nvpair_print_html_footer(stream,html_headers);
		fprintf(stream,"</center>\n");
	}

	fclose(stream);
}

static void show_version( const char *cmd )
{
        printf("%s version %d.%d.%d built by %s@%s on %s at %s\n",cmd,DTTOOLS_VERSION_MAJOR,DTTOOLS_VERSION_MINOR,DTTOOLS_VERSION_MICRO,BUILD_USER,BUILD_HOST,__DATE__,__TIME__);
}

static void show_help( const char *cmd )
{
	fprintf(stderr,"Use: %s [options]\n",cmd);
	fprintf(stderr,"   -p <port>      Port number to listen on (default is %d)\n",port);
	fprintf(stderr,"   -l <secs>      Lifetime of data, in seconds (default is %d)\n",lifetime);
	fprintf(stderr,"   -d <subsystem> Enable debugging for this subsystem\n");
	fprintf(stderr,"   -v             Show version string\n");
	fprintf(stderr,"   -h             Show this help screen\n");
}

int main( int argc, char *argv[] )
{
	struct datagram *update_port=0;
	struct link *link, *list_port=0;
	char ch;

	debug_config(argv[0]);

	while((ch=getopt(argc,argv,"p:l:d:hv"))!=(char)-1) {
		switch(ch) {
			case 'd':
				debug_flags_set(optarg);
				break;
			case 'p':
				port = atoi(optarg);
				break;
			case 'l':
				lifetime = atoi(optarg);
				break;
			case 'v':
				show_version(argv[0]);
				return 0;
			case 'h':
			default:
				show_help(argv[0]);
				return 1;
		}
	}

	signal(SIGPIPE,SIG_IGN);
	signal(SIGHUP,SIG_IGN);
	signal(SIGCHLD,SIG_DFL);

	domain_name_cache_guess(hostname);

	table = hash_cache_create(127,hash_string,(hash_cache_cleanup_t)nvpair_delete);
	if(!table) fatal("couldn't make hash table");
	
	update_port = datagram_create( port );
	if(!update_port) fatal("couldn't listen on udp port %d",port);

	list_port = link_serve( port );
	if(!list_port) fatal("couldn't listen on tcp port %d",port);

	while(1) {
		link = link_accept(list_port,5);
		if(link) {
			pid_t pid = fork();
			if(pid==0) {
				handle_query(link);
				exit(0);
			}
			link_close(link);
		}
		handle_updates(update_port);
	}

	return 1;
}


