/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "datagram.h"
#include "debug.h"
#include "domain_name_cache.h"
#include "stringtools.h"
#include "rotating_file.h"
#include "username.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <signal.h>

#include "getopt.h"

static void stop_server( int sig )
{
	exit(0);
}

static void show_version( const char *cmd )
{
	printf("%s %s %s\n",cmd,__DATE__,__TIME__);
}

static void show_use( const char *cmd ) 
{
	fprintf(stderr,"Use: %s [options]\n",cmd);
	fprintf(stderr,"Where options are:\n");
	fprintf(stderr," -o <file>  Send messages to this file.\n");
	fprintf(stderr," -O <bytes> Limit file to this size.\n");
	fprintf(stderr," -n         Do not perform reverse name lookups.\n");
	fprintf(stderr," -u <user>  Run as this userid.\n");
	fprintf(stderr," -v         Show the version string\n");
	fprintf(stderr," -h         Show this help screen.\n");
}

int main( int argc, char *argv[] )
{
	char buffer[DATAGRAM_PAYLOAD_MAX];
	char addr[DATAGRAM_ADDRESS_MAX];
	char name[DOMAIN_NAME_MAX];
	char  c;
	char *username=0;
	struct datagram *d;
	int port = DEBUG_PORT;
	int do_name = 1;

	while((c=getopt(argc,argv,"o:O:nu:vh"))!=(char)-1) {
		switch(c) {
			case 'o':
				debug_config_file(optarg);
				break;
			case 'O':
				debug_config_file_size(atoi(optarg));
				break;
			case 'u':
				username = optarg;
				break;
			case 'n':
				do_name = 0;
				break;
			case 'v':
				show_version(argv[0]);
				return 1;
			case 'h':
			default:
				show_use(argv[0]);
				return 1;
				break;
		}
	}

	signal(SIGPIPE,SIG_IGN);
	signal(SIGHUP,SIG_IGN);
	signal(SIGQUIT,stop_server);
	signal(SIGTERM,stop_server);

	if(username) {
		if(!username_set(username)) {
			fatal("Couldn't switch to user %s!",username);
		}
	} else {
		if(username_is_super()) {
			fatal("I refuse to run as the superuser.  Please use the -u option.");
		}
	}

	d = datagram_create( port );
	if(!d) fatal("couldn't listen on port %d: %s",port,strerror(errno));

	debug(D_RAW,"%s started.",argv[0]);

	while(1) {
		int result;

		result = datagram_recv(d,buffer,sizeof(buffer),addr,&port,60);
		if(result<0) continue;

		buffer[result] = 0;

		if(do_name) {
			domain_name_cache_lookup_reverse(addr,name);
		} else {
			strcpy(addr,name);
		}

		debug(D_RAW,"%s: %s",name,buffer);
	}
}
