/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "pfs_table.h"
#include "pfs_service.h"

extern "C" {
#include "chirp_reli.h"
#include "stringtools.h"
#include "debug.h"
}

#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <utime.h>

extern int pfs_master_timeout;

class pfs_file_chirp : public pfs_file
{
private:
	struct chirp_file *file;

public:
	pfs_file_chirp( pfs_name *name, struct chirp_file *f ) : pfs_file(name) {
		file = f;
	}

	virtual int close() {
		int result;
		debug(D_CHIRP,"close 0x%x",file);
		result = chirp_reli_close(file,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual pfs_ssize_t read( void *data, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t result;
		debug(D_CHIRP,"pread 0x%x 0x%x %d %d",file,data,length,offset);
		result = chirp_reli_pread(file,data,length,offset,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual pfs_ssize_t write( const void *data, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t result;
		debug(D_CHIRP,"pwrite 0x%x 0x%x %d %d",file,data,length,offset);
		result = chirp_reli_pwrite(file,data,length,offset,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int fstat( struct stat *buf ) {
		int result;
		debug(D_CHIRP,"fstat 0x%x",file);
		result = chirp_reli_fstat(file,buf,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int fstatfs( struct statfs *buf ) {
		int result;
		debug(D_CHIRP,"fstatfs 0x%x",file);
		result = chirp_reli_fstatfs(file,buf,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int ftruncate( pfs_size_t length ) {
		int result;
		debug(D_CHIRP,"ftruncate 0x%x %d",file,length);
		result = chirp_reli_ftruncate(file,length,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int fchmod( mode_t mode ) {
		int result;
		debug(D_CHIRP,"fchmod 0x%x %d",file,mode);
		result = chirp_reli_fchmod(file,mode,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int fchown( uid_t uid, gid_t gid ) {
		int result;
		debug(D_CHIRP,"fchown 0x%x %d %d",file,uid,gid);
		result = chirp_reli_fchown(file,uid,gid,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int fsync() {
		/* sync not needed because of synchronous writes */
		return 0;
	}

	virtual pfs_ssize_t get_size() {
		struct stat buf;
		if(this->fstat(&buf)==0) {
			return buf.st_size;
		} else {
			return -1;
		}
	}
};

static void add_to_dir( const char *path, void *arg )
{
	pfs_dir *dir = (pfs_dir *)arg;
	dir->append(path);
}

class pfs_service_chirp : public pfs_service {
public:
	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		struct chirp_file *file;		
		/* need synchronous writes because of recovery code */
		flags |= O_SYNC;
		debug(D_CHIRP,"open %s %d %d",name->rest,flags,mode);
		file = chirp_reli_open(name->host,name->rest,flags,mode,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= 0x%x %s",file,file ? "" : strerror(errno));
		if(file) {
			return new pfs_file_chirp(name,file);
		} else {
			return 0;
		}
	}

	virtual pfs_dir * getdir( pfs_name *name ) {
		int result=-1;
		pfs_dir *dir = new pfs_dir(name);
		debug(D_CHIRP,"getdir %s",name->rest);
		result = chirp_reli_getdir(name->host,name->rest,add_to_dir,dir,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,result>=0 ? "" : strerror(errno));
		if(result>=0) {
			return dir;
		} else {
			delete dir;
			return 0;
		}
	}

	virtual int stat( pfs_name *name, struct stat *buf ) {
		int result=-1;
		debug(D_CHIRP,"stat %s 0x%x",name->rest,buf);
		result = chirp_reli_stat(name->host,name->rest,buf,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int lstat( pfs_name *name, struct stat *buf ) {
		int result=-1;
		debug(D_CHIRP,"lstat %s 0x%x",name->rest,buf);
		result = chirp_reli_lstat(name->host,name->rest,buf,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int unlink( pfs_name *name ) {
		int result=-1;
		debug(D_CHIRP,"unlink %s",name->rest);
		result = chirp_reli_unlink(name->host,name->rest,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int access( pfs_name *name, mode_t mode ) {
		int result=-1;
		debug(D_CHIRP,"access %s %d",name->rest,mode);
		result = chirp_reli_access(name->host,name->rest,mode,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int chmod( pfs_name *name, mode_t mode ) {
		int result=-1;
		debug(D_CHIRP,"chmod %s %d",name->rest,mode);
		result = chirp_reli_chmod(name->host,name->rest,mode,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int chown( pfs_name *name, uid_t uid, gid_t gid ) {
		int result=-1;
		debug(D_CHIRP,"chown %s %d %d",name->rest,uid,gid);
		result = chirp_reli_chown(name->host,name->rest,uid,gid,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int lchown( pfs_name *name, uid_t uid, gid_t gid ) {
		int result=-1;
		debug(D_CHIRP,"lchown %s %d %d",name->rest,uid,gid);
		result = chirp_reli_lchown(name->host,name->rest,uid,gid,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int truncate( pfs_name *name, pfs_off_t length ) {
		int result=-1;
		debug(D_CHIRP,"truncate %s %d",name->rest,length);
		result = chirp_reli_truncate(name->host,name->rest,length,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int utime( pfs_name *name, struct utimbuf *t ) {
		int result=-1;
		debug(D_CHIRP,"utime %s %d %d",name->rest,t->actime,t->modtime);
		result = chirp_reli_utime(name->host,name->rest,t->actime,t->modtime,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int rename( pfs_name *name, pfs_name *newname ) {
		int result=-1;
		debug(D_CHIRP,"rename %s %s",name->rest,newname->rest);
		result = chirp_reli_rename(name->host,name->rest,newname->rest,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int chdir( pfs_name *name, char *newname ) {
		int result=-1;
		char tmp[PFS_PATH_MAX];
		debug(D_CHIRP,"canonicalize %s %d",name->rest,sizeof(tmp));
		result = chirp_reli_canonicalize(name->host,name->rest,tmp,sizeof(tmp),time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		if(result>=0) {
			tmp[result] = 0;
			sprintf(newname,"/%s/%s:%d%s",name->service_name,name->host,name->port,tmp);
			result = 0;
		}	
		return result;
	}

	virtual int link( pfs_name *name, pfs_name *newname ) {
		int result=-1;
		debug(D_CHIRP,"link %s %s",name->rest,newname->rest);
		result = chirp_reli_link(name->host,name->rest,newname->rest,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int symlink( pfs_name *name, pfs_name *newname ) {
		int result=-1;
		debug(D_CHIRP,"symlink %s %s",name->rest,newname->rest);
		result = chirp_reli_symlink(name->host,name->rest,newname->rest,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}
	virtual int readlink( pfs_name *name, char *buf, pfs_size_t length ) {
		int result=-1;
		debug(D_CHIRP,"readlink %s 0x%x %d",name->rest,buf,length);
		result = chirp_reli_readlink(name->host,name->rest,buf,length,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		/* Fix up relative-rooted paths to match our view of things */
		if(result>0 && buf[0]=='/') {
			char tmp[PFS_PATH_MAX];
			buf[result] = 0;
			strcpy(tmp,buf);
			sprintf(buf,"/%s/%s%s",name->service_name,name->host,tmp);
			result = strlen(buf);
		}
		return result;
	}

	virtual int mkdir( pfs_name *name, mode_t mode ) {
		int result=-1;
		debug(D_CHIRP,"mkdir %s %d",name->rest,mode);
		result = chirp_reli_mkdir(name->host,name->rest,mode,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int rmdir( pfs_name *name ) {
		int result=-1;
		debug(D_CHIRP,"rmdir %s",name->rest);
		result = chirp_reli_rmdir(name->host,name->rest,time(0)+pfs_master_timeout);
		debug(D_CHIRP,"= %d %s",result,((result>=0) ? "" : strerror(errno)) );
		return result;
	}

	virtual int is_seekable() {
		return 1;
	}

};

static pfs_service_chirp pfs_service_chirp_instance;
pfs_service *pfs_service_chirp = &pfs_service_chirp_instance;


