/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#ifdef HAS_KANGAROO

#include "pfs_table.h"
#include "pfs_service.h"

extern "C" {
#include "debug.h"
#include "kangaroo_client.h"
#include "stringtools.h"
}

#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <errno.h>
#include <stdlib.h>

static struct kangaroo_server *server=0;
static int did_write=0;

static int kangaroo_init()
{
	if(!server) server = kangaroo_server_open(0,0);
	if(server) {
		return 1;
	} else {
		return 0;
	}
}

int pfs_service_kangaroo_cleanup()
{
	if(server && did_write) {
		debug(D_KANGAROO,"commit");
		if(kangaroo_commit(server)!=0) {
			fatal("couldn't safely commit output: %s\n",strerror(errno));
		}
		debug(D_KANGAROO,"= Success");
	}
	return 0;
}

static int kangaroo_file_exists( pfs_name *name )
{
	int result;
	char c;

	if(!kangaroo_init()) return 0;

	debug(D_KANGAROO,"get %s %s 0 1",name->host,name->rest);
	result = kangaroo_get(server,name->host,name->rest,0,1,&c);
	debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
	if(result>=0) {
		return 1;
	} else {
		return 0;
	}
}

class pfs_file_kangaroo : public pfs_file
{
public:
	pfs_file_kangaroo(pfs_name *name) : pfs_file(name) {
	}

	virtual int close() {
		return 0;
	}

	virtual pfs_ssize_t read( void *data, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t result;
		debug(D_KANGAROO,"get %s %s %d %d",name.host,name.rest,offset,length);
		result = kangaroo_get(server,name.host,name.rest,offset,length,data);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}

	virtual pfs_ssize_t write( const void *data, pfs_size_t length, pfs_off_t offset ) {
		pfs_ssize_t result;
		did_write=1;
		debug(D_KANGAROO,"put %s %s %d %d",name.host,name.rest,offset,length);
		result = kangaroo_put(server,name.host,name.rest,offset,length,data);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}

	virtual int fsync() {
		int result;
		debug(D_KANGAROO,"push %s %s",name.host,name.rest);
		result = kangaroo_push(server,name.host,name.rest);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}

	virtual int fstat( struct stat *buf ) {
		pfs_service_emulate_stat(&name,buf);
		return 0;
	}

	virtual int ftruncate( pfs_size_t length ) {
		int result;
		debug(D_KANGAROO,"truncate %s %s %d",name.host,name.rest,length);
		result = kangaroo_truncate(server,name.host,name.rest,length);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}
};

class pfs_service_kangaroo : public pfs_service {
public:
	virtual ~pfs_service_kangaroo() {
	}

	virtual pfs_file * open( pfs_name *name, int flags, mode_t mode ) {
		int result;

		if(!kangaroo_init()) return 0;

		if(flags&O_CREAT) {
			debug(D_KANGAROO,"create %s %s %d",name->host,name->rest,mode);
			result = kangaroo_create(server,name->host,name->rest,mode);
			debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		}

		if(flags&O_TRUNC) {
			debug(D_KANGAROO,"truncate %s %s %d",name->host,name->rest,0);
			result = kangaroo_truncate(server,name->host,name->rest,0);
			debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		}

		return new pfs_file_kangaroo(name);
	}

	virtual int stat( pfs_name *name, struct stat *buf ) {
		if(kangaroo_file_exists(name)) {
			pfs_service_emulate_stat(name,buf);
			return 0;
		} else {
			errno = ENOENT;
			return -1;
		}
	}

	virtual int lstat( pfs_name *name, struct stat *buf ) {
		return stat(name,buf);
	}

	virtual int access( pfs_name *name, mode_t mode ) {
		if( mode&R_OK || mode&F_OK ) {
			if(kangaroo_file_exists(name)) {
				return 0;
			} else {
				errno = ENOENT;
				return -1;
			}
		} else {
			return 0;
		}
		return 0;
	}

	virtual int truncate( pfs_name *name, pfs_off_t length ) {
		int result;
		if(!kangaroo_init()) return -1;
		debug(D_KANGAROO,"truncate %s %s %d",name->host,name->rest,0);
		result = kangaroo_truncate(server,name->host,name->rest,length);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}

	virtual int unlink( pfs_name *name ) {
		int result;
		if(!kangaroo_init()) return -1;
		debug(D_KANGAROO,"delete %s %s",name->host,name->rest);
		result = kangaroo_delete(server,name->host,name->rest);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}

	virtual int rename( pfs_name *name, pfs_name *name2) {
		int result;
		if(strcmp(name->host,name2->host)) {
			errno = EXDEV;
			return -1;
		}
		if(!kangaroo_init()) return -1;
		debug(D_KANGAROO,"rename %s %s %s",name->host,name->rest,name2->rest);
		result = kangaroo_rename(server,name->host,name->rest,name2->rest);
		debug(D_KANGAROO,"= %d %s",result,result>=0 ? "" : strerror(errno));
		return result;
	}

	virtual int is_seekable() {
		return 1;
	}

};

static pfs_service_kangaroo pfs_service_kangaroo_instance;
pfs_service *pfs_service_kangaroo = &pfs_service_kangaroo_instance;

#endif
