/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#ifndef PFS_SYSDEPS_H
#define PFS_SYSDEPS_H

/*
In this file, we establish all of the definitions
and feature activations that are dependent upon
each system, with some explanation as to why they
are necessary.

Notice that we simply define our own private
versions of structures like stat, stat64, and so forth.
These are *not* the user level versions of these
structure, *nor* the kernel level version.  These are
the structures used at the kernel interface, which
are occasionally different than the other two, and
remarkably difficult to pull in a definition from
the right include files.  So, we just define our own.
*/

#ifndef _LARGEFILE64_SOURCE
#define _LARGEFILE64_SOURCE
#endif

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "pfs_types.h"

#if defined(linux)

#include <sched.h>
#include <linux/net.h>

/*
On Linux,the arguments to mmap come indirectly
through a structure pointed to by the first arg.
*/

#define PFS_INDIRECT_MMAP_ARGS 1

/*
And, with mmap2, the offset in the file is given
in integral pages, not bytes
*/

#define PFS_MMAP_OFFSET_IN_PAGES 1

/*
On Linux, ptrace forces debugged processes to become
children of those that debug them.  Thus, we must keep
track of the process trees and deliver process completion
signals ourself.
*/

#define PFS_MANAGE_PROCESS_TREES 1

/*
On Linux, fork() does not cause the child process to
inherit the ptrace attribute.  Instead, we must
convert fork() into an equivalent clone() with the
ptrace attribute.  Further, the necessary flags
are not always defined in all kernels.
*/

#define PFS_CONVERT_FORK_TO_CLONE 1

#ifndef CLONE_PTRACE
#define CLONE_PTRACE    0x00002000
#endif

#ifndef CLONE_PARENT
#define CLONE_PARENT    0x00008000
#endif

/*
We require a scratch area in each process' address
space for storing internal data such as an adjusted
argument to execve.  This need not be *free* space,
as we will restore its contents afterwards.  It must
simply be a valid memory address.
*/

#define PFS_SCRATCH_ADDR ((char*)0x08048000)
#define PFS_SCRATCH_SIZE 4096

/*
Sadly, it is necessary to prefix each of the structure
tags with pfs (i.e. pfs_st_dev instead of st_dev) because
the structure tags themselves are frequently the target
of a macro definition that munges something at the user
level that does not concern us.
*/

struct pfs_kernel_stat {
	unsigned long pfs_st_dev;
	unsigned long pfs_st_ino;
	unsigned short pfs_st_mode;
	unsigned short pfs_st_nlink;
	unsigned short pfs_st_uid;
	unsigned short pfs_st_gid;
	unsigned long pfs_st_rdev;
	unsigned long pfs_st_size;
	unsigned long pfs_st_blksize;
	unsigned long pfs_st_blocks;
	unsigned long pfs_st_atime;
	unsigned long pfs_st_p1;
	unsigned long pfs_st_mtime;
	unsigned long pfs_st_p2;
	unsigned long pfs_st_ctime;
	unsigned long pfs_st_p3;
	unsigned long pfs_st_p4;
	unsigned long pfs_st_p5;
};

struct pfs_kernel_stat64 {
	unsigned long long pfs_st_dev;
	unsigned long pfs_st_pad1;
	unsigned long pfs_st_ino;
	unsigned long pfs_st_mode;
	unsigned long pfs_st_nlink;
	unsigned long pfs_st_uid;
	unsigned long pfs_st_gid;
	unsigned long long pfs_st_rdev;
	unsigned long pfs_st_pad2;
	unsigned long long pfs_st_size;
	unsigned long pfs_st_blksize;
	unsigned long long pfs_st_blocks;
	unsigned long pfs_st_atime;
	unsigned long pfs_st_unused1;
	unsigned long pfs_st_mtime;
	unsigned long pfs_st_unused2;
	unsigned long pfs_st_ctime;
	unsigned long pfs_st_unused3;
	unsigned long long pfs_st_ino_extra;
};

struct pfs_kernel_statfs {
        long pfs_f_type;
        long pfs_f_bsize;
        long pfs_f_blocks;
        long pfs_f_bfree;
        long pfs_f_bavail;
        long pfs_f_files;
        long pfs_f_ffree;
	fsid_t pfs_f_fsid;
        long pfs_f_namelen;
        long pfs_f_spare[6];
};

/* Linux has no statfs64 at the kernel interface yet! */

struct pfs_kernel_dirent {
        unsigned long   pfs_d_ino;
        unsigned long   pfs_d_off;
        unsigned short  pfs_d_reclen;
	unsigned char   pfs_d_type;
        char            pfs_d_name[PFS_PATH_MAX];
};

struct pfs_kernel_dirent64 {
        unsigned long long pfs_d_ino;
        unsigned long long pfs_d_off;
        unsigned short     pfs_d_reclen;
	unsigned char      pfs_d_type;
        char               pfs_d_name[PFS_PATH_MAX];
};

#else
#error This platform is not supported yet!
#endif

/*
Many data structures must be aligned on 8 byte boundaries.
This rounds up values to multiples of 8.
*/

#ifndef _ROUND_UP
#define _ROUND_UP(x,n) (((x)+(n)-1u) & ~((n)-1u))
#endif

#ifndef ROUND_UP
#define ROUND_UP(x) _ROUND_UP(x,8LL)
#endif

/*
Handy for copying one structure to another.
*/

#define COPY_STAT( a, b ) \
	memset(&(b),0,sizeof((b)));\
	(b).pfs_st_dev = (a).st_dev;\
	(b).pfs_st_ino = (a).st_ino;\
	(b).pfs_st_mode = (a).st_mode;\
	(b).pfs_st_nlink = (a).st_nlink;\
	(b).pfs_st_uid = (a).st_uid;\
	(b).pfs_st_gid = (a).st_gid;\
	(b).pfs_st_rdev = (a).st_rdev;\
	(b).pfs_st_size = (a).st_size;\
	(b).pfs_st_blksize = (a).st_blksize;\
	(b).pfs_st_blocks = (a).st_blocks;\
	(b).pfs_st_atime = (a).st_atime;\
	(b).pfs_st_mtime = (a).st_mtime;\
	(b).pfs_st_ctime = (a).st_ctime;

#define COPY_STATFS( a, b ) \
	memset(&(b),0,sizeof(b));\
        (b).pfs_f_type = (a).f_type;\
        (b).pfs_f_bsize = (a).f_bsize;\
        (b).pfs_f_blocks = (a).f_blocks;\
        (b).pfs_f_bfree = (a).f_bfree;\
        (b).pfs_f_bavail = (a).f_bavail;\
        (b).pfs_f_files = (a).f_files;\
        (b).pfs_f_ffree = (a).f_ffree;\
	(b).pfs_f_fsid = (a).f_fsid;\
        (b).pfs_f_namelen = (a).f_namelen;

/*
The size of a dirent is the size of the structure
without the name field, plus the actual length
of the null-terminated name, rounded up to 8-byte alignment.
*/

#define DIRENT_SIZE( x ) \
	ROUND_UP(((char*)&(x).pfs_d_name[0]-(char*)&(x)) + strlen((x).pfs_d_name) + 1)

#define COPY_DIRENT( a, b ) \
	memset(&(b),0,sizeof((b))); \
	strcpy((b).pfs_d_name,(a).d_name);\
	(b).pfs_d_ino = (a).d_ino;\
	(b).pfs_d_off = (a).d_off;\
	(b).pfs_d_reclen = DIRENT_SIZE(b);

#endif
