/*
Copyright (C) 2004 Douglas Thain
This work is made available under a BSD-style license.
See the file COPYING for details.
*/

#include "tracer.h"
#include "stringtools.h"
#include "full_io.h"
#include "xmalloc.h"
#include "debug.h"

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <fcntl.h>
#include <limits.h>

#include <sys/wait.h>
#include <sys/ptrace.h>

#include <linux/user.h>

struct tracer {
	pid_t pid;
	int memory_file;
	int gotregs;
	struct user_regs_struct regs;
};

void tracer_prepare()
{
	ptrace(PTRACE_TRACEME,0,0,0);
	/* Ignore return value, as it is sometimes bogus. */
}

struct tracer * tracer_attach( pid_t pid )
{
	struct tracer *t;
	char path[PATH_MAX];

	t = malloc(sizeof(*t));
	if(!t) return 0;

	t->pid = pid;
	t->gotregs = 0;

	sprintf(path,"/proc/%d/mem",pid);
	t->memory_file = open(path,O_RDWR);
	if(t->memory_file<0) {
		free(t);
		return 0;
	}

	return t;
}

void tracer_detach( struct tracer *t )
{
	ptrace(PTRACE_DETACH,t->pid,0,0);
	close(t->memory_file);
	free(t);
}

void tracer_continue( struct tracer *t, int signum )
{
	ptrace(PTRACE_SYSCALL,t->pid,0,signum);
	t->gotregs = 0;
}

int tracer_args_get( struct tracer *t, int *syscall, int args[TRACER_ARGS_MAX] )
{
	if(!t->gotregs) {
		if(ptrace(PTRACE_GETREGS,t->pid,0,&t->regs)!=0) return 0;
		t->gotregs = 1;
	}
	*syscall = t->regs.orig_eax;
	args[0] = t->regs.ebx;
	args[1] = t->regs.ecx;
	args[2] = t->regs.edx;
	args[3] = t->regs.esi;
	args[4] = t->regs.edi;
	args[5] = t->regs.ebp;

	return 1;
}

int tracer_args_set( struct tracer *t, int syscall, int args[TRACER_ARGS_MAX], int nargs )
{
	if(!t->gotregs) {
		if(ptrace(PTRACE_GETREGS,t->pid,0,&t->regs)!=0) return 0;
		t->gotregs = 1;
	}
	t->regs.orig_eax = syscall;
	if(nargs>=1) t->regs.ebx = args[0];
	if(nargs>=2) t->regs.ecx = args[1];
	if(nargs>=3) t->regs.edx = args[2];
	if(nargs>=4) t->regs.esi = args[3];
	if(nargs>=5) t->regs.edi = args[4];
	if(nargs>=6) t->regs.ebp = args[5];
	if(ptrace(PTRACE_SETREGS,t->pid,0,&t->regs)!=0) return 0;
	return 1;
}

int tracer_result_get( struct tracer *t, int *result )
{
	if(!t->gotregs) {
		if(ptrace(PTRACE_GETREGS,t->pid,0,&t->regs)!=0) return 0;
		t->gotregs = 1;
	}
	*result = t->regs.eax;
	return 1;
}

int tracer_result_set( struct tracer *t, int result )
{
	if(!t->gotregs) {
		if(ptrace(PTRACE_GETREGS,t->pid,0,&t->regs)!=0) return 0;
		t->gotregs = 1;
	}
	t->regs.eax = result;
	if(ptrace(PTRACE_SETREGS,t->pid,0,&t->regs)!=0) return 0;
	return 1;
}

static int tracer_copy_out_slow( struct tracer *t, const void *data, const void *uaddr, int length )
{
	unsigned long *ldata = (unsigned long*)data;
	unsigned long *luaddr = (unsigned long*)uaddr;
	int size = length;

	/* first, copy whole words */

	while(size>=sizeof(long)) {
		ptrace(PTRACE_POKEDATA,t->pid,luaddr,*ldata);
		size -= sizeof(long);
		luaddr++;
		ldata++;
	}

	/* if necessary, copy the last few bytes */

	if(size>0) {
		long temp;
		temp = ptrace(PTRACE_PEEKDATA,t->pid,luaddr,0);
		memcpy(&temp,ldata,size);
		ptrace(PTRACE_POKEDATA,t->pid,luaddr,temp);
	}

	return length;		
}

int tracer_copy_out( struct tracer *t, const void *data, const void *uaddr, int length )
{
	int result;
	static int has_fast_write=1;

	if(length==0) return 0;

	if(has_fast_write) {
		result = full_pwrite(t->memory_file,data,length,(int)uaddr);
		if( result!=length && errno==EINVAL ) {
			has_fast_write = 0;
		} else {
			return result;
		}
	}
	result = tracer_copy_out_slow(t,data,uaddr,length);
	return result;
}

/*
Tidbit:
The reason we must have pread64 here is that the target address
of the read in user space can easily be above 2GB (2^31) on 
many machines.  We don't need full 64 bit access, but we need
to access addresses above 2GB, which can only be done with the
64-bit calls.
*/

int tracer_copy_in( struct tracer *t, void *data, const void *uaddr, int length )
{
	return full_pread64(t->memory_file,data,length,(INT64_T)(unsigned)uaddr);
}

#include "tracer.table.c"

const char * tracer_syscall_name( struct tracer *t, int syscall )
{
	if( syscall<0 || syscall>SYSCALL_MAX ) {
		return "unknown";
	} else {
		return syscall_names[syscall];
	}
}
